@@ -27,9 +27,8 @@
 #include "pulseaudio/qpulsehelpers_p.h"
 
 #include <private/qaudiosystem_p.h>
-
+#include <private/qaudiostatemachine_p.h>
 #include <pulse/pulseaudio.h>
-#include <atomic>
 
 QT_BEGIN_NAMESPACE
 
@@ -40,7 +39,7 @@ class QPulseAudioSink : public QPlatformAudioSink
 
 public:
     QPulseAudioSink(const QByteArray &device, QObject *parent);
-    ~QPulseAudioSink();
+    ~QPulseAudioSink() override;
 
     void start(QIODevice *device) override;
     QIODevice *start() override;
@@ -67,8 +66,8 @@ protected:
     void timerEvent(QTimerEvent *event) override;
 
 private:
-    void setStateAndError(QAudio::State state, QAudio::Error error, bool forceEmitState = false);
-    void startReading();
+    void startPulling();
+    void stopTimer();
 
     bool open();
     void close();
@@ -78,9 +77,13 @@ private Q_SLOTS:
     void userFeed();
     void onPulseContextFailed();
 
-    PAOperationUPtr exchangeDrainOperation(pa_operation *newOperation);
-
 private:
+    using PAOperationHandle = QPulseAudioInternal::PAOperationHandle;
+    using PAStreamHandle = QPulseAudioInternal::PAStreamHandle;
+    PAOperationHandle exchangeDrainOperation(pa_operation *newOperation);
+
+    qsizetype defaultBufferSize() const;
+
     pa_sample_spec m_spec = {};
     // calculate timing manually, as pulseaudio doesn't give us good enough data
     mutable timeval lastTimingInfo = {};
@@ -93,7 +96,7 @@ private:
     QBasicTimer m_tickTimer;
 
     QIODevice *m_audioSource = nullptr;
-    pa_stream *m_stream = nullptr;
+    PAStreamHandle m_stream;
     std::vector<char> m_audioBuffer;
 
     qint64 m_totalTimeValue = 0;
@@ -102,16 +105,15 @@ private:
     mutable qint64 lastProcessedUSecs = 0;
     qreal m_volume = 1.0;
 
-    std::atomic<QAudio::Error> m_errorState = QAudio::NoError;
-    std::atomic<QAudio::State> m_deviceState = QAudio::StoppedState;
-    QAudio::State m_suspendedInState = QAudio::SuspendedState;
     std::atomic<pa_operation *> m_drainOperation = nullptr;
-    int m_periodSize = 0;
-    int m_bufferSize = 0;
-    int m_periodTime = 0;
+    qsizetype m_bufferSize = 0;
+    std::optional<qsizetype> m_userBufferSize = std::nullopt;
+    int m_pullingPeriodSize = 0;
+    int m_pullingPeriodTime = 0;
     bool m_pullMode = true;
     bool m_opened = false;
-    bool m_resuming = false;
+
+    QAudioStateMachine m_stateMachine;
 };
 
 class PulseOutputPrivate : public QIODevice
@@ -121,7 +123,7 @@ class PulseOutputPrivate : public QIODevice
 
 public:
     PulseOutputPrivate(QPulseAudioSink *audio);
-    virtual ~PulseOutputPrivate() {}
+    ~PulseOutputPrivate() override = default;
 
 protected:
     qint64 readData(char *data, qint64 len) override;
